from defichainUtils.utils import ParamType
from defichainUtils.utils import utils,CustomTxDecode
import defichainUtils.utils.chain as chain
from defichainUtils.utils import CustomTx
# from utils import ParamType
# from utils import utils,CustomTxDecode
# import utils.chain as chain
# from utils import CustomTx
from dataclasses import dataclass


VERSION = 4
SEQUENCE = 'ffffffff'
DECIMALS = 100000000
def compose(txs:list):
    '''
    List of DfTx Class objects. Can be used for output (2.) in createrawtransaction (RPC DefiChain) 
    '''
    res = []
    for tx in txs:
        res.append(tx.hex())
    return res

def createRawTransaction(input:list,output:list,shareAddress:str,locktime:int=0,replaceable:bool=False):
    '''
    For Witness Transactions and BECH32 Addresses Only!!

    input transactions is list of dictionaries and need fields: txid, vout, amount

    output list are hex endcoded transactions. 
    UTXO output is calculated automatically and amount minus fee is sent to shareAddress.
    '''
    amount = 0
    res = ''
    res = res + utils.int2hex(VERSION,'little',4)
    res = res + utils.int2hex(len(input),'little')
    # (Unspent) UTXO
    for i in input:
        res = res + utils.convert_hex(i['txid'],'big','little')
        res = res + utils.int2hex(int(i['vout']),'little',4)
        res = res + '00'
        res = res + utils.generateSequence(SEQUENCE,locktime,replaceable)
        amount = amount + float(i['amount'])
    
    res = res + utils.int2hex(len(output)+1,'little')
    # OP_RETURN Transactions
    for i,o in enumerate(output):
        # '00' is seperator between transactions
        if i > 0:
            res = res + '00'
        res = res + utils.int2hex(0,'little',8)
        res = res + utils.int2hex(int(len(o)/2),'little')
        res = res + o

    # UTXO Output
    # '00' is seperator between transactions
    rbf = 0
    if replaceable:
        rbf = 1
    
    # TODO: signrawtransaction and get number of bytes to calculate vSize.
    dummy = '00' + utils.int2hex(0,'little',8) + utils.encodeBech32AddressToHex(shareAddress) + utils.int2hex(rbf,'little',1) + utils.int2hex(locktime,'little',4)
    
    size = int(len(res+dummy)/2)
    #size = int(736/2)
    size = int(580/2)
    # Witness data is 108 Bytes; vsize is "size" plus "3 time size minus witness data" divided by four. https://en.bitcoinwiki.org/wiki/Block_weight#Detailed_example ; https://bitcointalk.org/index.php?topic=5276203.0
    vSize = size - (108 * 3 / 4)
    print(utils.int2hex(int(amount*DECIMALS - vSize),'little',8))
    outUTXO = '00' + utils.int2hex(int(amount*DECIMALS - vSize),'little',8) + utils.encodeBech32AddressToHex(shareAddress) + utils.int2hex(rbf,'little',1) + utils.int2hex(locktime,'little',4)
    return res + outUTXO

def decodeCustomTx(hex:str):
    expectedOPReturn = hex[:2]
    
    hexLength=hex[2:4]
    if hexLength == '4c':
        hexLength = hexLength=hex[4:6]
        DfTxStart = 6
    elif hexLength == '4d':
        hexLength = utils.convert_hex(hex[4:8],'little','big')
        DfTxStart = 8
    else:
        DfTxStart = 4
    expectedDfMarker = hex[DfTxStart:DfTxStart + 8]

    #hexLengthV2=hex[4:6]
    #expectedDfTxMarkerV2 = hex[6:14]
    if expectedOPReturn != '6a':
        # TODO: Raise Error
        return {
            'error': 1,
            'msg': 'Missing key word for OP_RETURN'
        }
    if expectedDfMarker == utils.stringToHex('DfTx'):
        hex = hex[DfTxStart + 8:] 
    elif expectedDfMarker == utils.stringToHex('DfAf'):
        hex = hex[DfTxStart + 8:] 
    elif expectedDfMarker == utils.stringToHex('DfAP'):
        hex = hex[DfTxStart + 8:] 
    elif expectedDfMarker == utils.stringToHex('DfTS'):
        hex = hex[DfTxStart + 8:] 
    else:
        # TODO: Raise Error
        return {
            'error': 1,
            'msg': 'Missing key word for Marker'
        }

    if int(hexLength,16) != int(len(hex)/2) + len('DfTx'):
        # TODO: Raise Error
        return {
            'error': 1,
            'msg': 'Indicated length in hex string does fit string length'
        }
    return CustomTxDecode.decodeCustomTx(hex,expectedDfMarker)

def processCustomTransactionOnChain(block,name,parameters:dict,pool:dict):
    '''
    block: blockHeight
    name: Name of the rpc function (lowercase), e.g. poolswap (compositeswap is a chain of poolswaps!!! Act accordingly. Only poolswap possible)
    pool: pool keys with value - "reserveA", "reserveB", "totalLiquidity", "poolSymbol" (e.g. TSLA-DUSD)
    parameters: function keys with value - 
        - addpoolliquidity: "amountA", "amountB" (amount of Tokens; must be in correct order depending on poolpair. E.g. TSLA-DUSD means amountA is TSLA Amount)
        - removePoolLiquidity: "liquidity" (amount of Poolpair Tokens)
        - poolSwap: "tokenFrom", "fromAmount", "commission", "dexFeeInPct"
    
    RETURN:
    addpoolliquidity: amount liquidity tokens
    removepoolliquidity: amount of TokenA and TokenB (tuple) ->
    poolswap: amount of new reserveFrom, new reserveTo and amount TokenTo (tuple)
    '''
    if 'reserveA' not in pool or 'reserveB' not in pool:
        # TODO: Raise Error
        return {
            'error': 1,
            'msg': 'reserveA and/or reserveB missing'
        }
    if name == 'addpoolliquidity':
        if 'amountA' not in parameters or 'amountB' not in parameters or 'totalLiquidity' not in pool:
            # TODO: Raise Error
            return {
                'error': 1,
                'msg': 'amountA and/or amountB and/or totalLiquidity missing'
            } 
        liquidity = chain.addPoolLiquidity(parameters['amountA'],parameters['amountB'],pool['reserveA'],pool['reserveB'],pool['totalLiquidity'])
        return {
            'success': 1,
            'liquidity': liquidity
        }
    elif name == 'removepoolliquidity':
        if 'liquidity' not in parameters or 'totalLiquidity' not in pool:
            # TODO: Raise Error
            return {
                'error': 1,
                'msg': 'liquidity and/or totalLiquidity missing'
            } 
        (amountA,amountB) = chain.removePoolLiquidity(parameters['liquidity'],pool['reserveA'],pool['reserveB'],pool['totalLiquidity'])
        return {
            'success': 1,
            'amountA': amountA,
            'amountB': amountB
        }
    elif name == 'poolswap':
        if 'fromAmount' not in parameters or 'isForward' not in parameters or 'commission' not in parameters or 'dexFeeInPct' not in parameters:
            # TODO: Raise Error
            return {
                'error': 1,
                'msg': 'tokenFrom and/or fromAmount and/or symbol and/or commission and/or dexFeeInPct missing'
            } 
        (reserveAChange,reserveBChange,swapped,commissionAmount,dexFeeInAmount,dexFeeOutAmount) = chain.poolSwap(block,parameters['fromAmount'],parameters['isForward'],pool['reserveA'],pool['reserveB'],parameters['commission'],parameters['dexFeeInPct'],parameters['dexFeeOutPct'])
        return {
            'success': 1,
            'reserveAChange': reserveAChange,
            'reserveBChange': reserveBChange,
            'swapped': swapped,
            'commissionAmount': commissionAmount,
            'dexFeeInAmount': dexFeeInAmount,
            'dexFeeOutAmount': dexFeeOutAmount
        }

# TODO: only working in python3.9 not in 3.8!!!
# class TakeLoan(CustomTx.DfTx):
#     def __init__(self,vaultId:str,address:str,amounts:list,rpcConnector):
#         super().__init__()
#         resAmounts = []
#         for a in amounts:
#             amount = a.split('@')
#             resAmounts.append((float(amount[0]),utils.getTokenId(amount[1],rpcConnector)))
#         self.params = [CustomTx.CustomTxType.TakeLoan,ParamType.Vault(vaultId),ParamType.Address(address),ParamType.Amounts(resAmounts)]

# class AddPoolLiquidity(CustomTx.DfTx):
#     def __init__(self,amountPerAddress:list[tuple[str,str]],shareAddress:str,rpcConnector):
#         super().__init__()
#         resAddressAmount = []
#         for a in amountPerAddress:
#             amount = a[1].split('@')
#             resAddressAmount.append((a[0],float(amount[0]),utils.getTokenId(amount[1],rpcConnector)))
#         self.params = [CustomTx.CustomTxType.AddPoolLiquidity,ParamType.AddressesWithAmount(resAddressAmount),ParamType.Address(shareAddress)]

# @dataclass
# class RPCConnector():
#     NODE_URL:str
#     NODE_USER:str
#     NODE_PASSWORD:str


# if __name__ == '__main__':
#     #a = 'df1qamj6pp30h9y3syqcz8nwnekzg8dswul3tw5kjy'
#     #adr = utils.encodeBech32AddressToHex(a)
#     #     6a2a44665478700700400d03000000000017a914f78ca7530bd35fff6af98a49522a34f7508ab64e87010000
#     hex = '6a1044665453010000003100000014000000'
#     #hex = '6a4c4f4466547869160014eee5a0862fb94918101811e6e9e6c241db0773f1000a00000000000000160014eee5a0862fb94918101811e6e9e6c241db0773f102ffffffffffffff7fffffffffffffff7f0105'
#     print(decodeCustomTx(hex))
    
    # print(hex(2059200))
    #print(utils.int2hex(144015528,'little',8))
    # #print(int(utils.convert_hex('b95f2f0300000000','little','big'),16)/100000000)
    # print(int(utils.convert_hex('90eac90100000000','big','little'),16))
    

    # # This is how this package could be used.
    # v = 'ed299afd57b7b354983efce9254436c7409e2c52a152d0a41bec0fb477f3a0b6'
    # a = 'tf1q6qj52ykxlf6halmx0g32gaumuuptactwgrqh23'

    # # a = 'df1q6qcutr37ex6xd5mjcmgrlr9239ck9z22605hfp'

    # RPCConn =RPCConnector("","","")
    # t = TakeLoan(v,a,["0.00112715@MSFT","1@DUSD"],RPCConn)
    # #print(t.hex())
    # p = AddPoolLiquidity([(a,"1@DUSD"),(a,"0.00112715@MSFT")],a,RPCConn)

    # input = {
    #     "txid":"b8639f6f27318c281f574ecb0360c1340e91e32780ca8da5c621a64a24026739",
    #     "vout":1,
    #     "amount":0.91485041
    # }
    # print(createRawTransaction([input],[t.hex()],a))
    # # print(t.hex())
    # # #d = DfTx(cTxType.TakeLoan,[v,a])
    # # #print(d.getHex())

    



import math
from typing import Optional

MINIMUM_LIQUIDITY = 1000
SLOPE_SWAP_RATE = 1000
# https://github.com/DeFiCh/ain/blob/master/src/chainparams.cpp
BAY_FRONT_GARDENS_HEIGHT = 488300
FORT_CANNING_HILL_HEIGHT = 1604999

class ArithUint256:
    # https://github.com/bitcoin/bitcoin/blob/master/src/arith_uint256.cpp

    __slots__ = '_value', '_compact'

    def __init__(self, value: int) -> None:
        self._value = value
        self._compact: Optional[int] = None

    @classmethod
    def from_compact(cls, compact) -> 'ArithUint256':
        size = compact >> 24
        word = compact & 0x007fffff
        if size <= 3:
            return cls(word >> 8 * (3 - size))
        else:
            return cls(word << 8 * (size - 3))

    @property
    def value(self) -> int:
        return self._value

    @property
    def compact(self) -> int:
        if self._compact is None:
            self._compact = self._calculate_compact()
        return self._compact

    @property
    def negative(self) -> int:
        return self._calculate_compact(negative=True)

    @property
    def bits(self) -> int:
        """ Returns the position of the highest bit set plus one. """
        bits = bin(self._value)[2:]
        for i, d in enumerate(bits):
            if d:
                return (len(bits) - i) + 1
        return 0

    @property
    def low64(self) -> int:
        return self._value & 0xffffffffffffffff

    def _calculate_compact(self, negative=False) -> int:
        size = (self.bits + 7) // 8
        if size <= 3:
            compact = self.low64 << 8 * (3 - size)
        else:
            compact = ArithUint256(self._value >> 8 * (size - 3)).low64
        # The 0x00800000 bit denotes the sign.
        # Thus, if it is already set, divide the mantissa by 256 and increase the exponent.
        if compact & 0x00800000:
            compact >>= 8
            size += 1
        assert (compact & ~0x007fffff) == 0
        assert size < 256
        compact |= size << 24
        if negative and compact & 0x007fffff:
            compact |= 0x00800000
        return compact

    def __mul__(self, x):
        # Take the mod because we are limited to an unsigned 256 bit number
        return ArithUint256((self._value * x) % 2 ** 256)
    
    def __add__(self, x):
        # Take the mod because we are limited to an unsigned 256 bit number
        return ArithUint256((self._value + x) % 2 ** 256)
    
    def __sub__(self, x):
        # Take the mod because we are limited to an unsigned 256 bit number
        return ArithUint256((self._value - x))

    def __truediv__(self, x):
        return ArithUint256(int(self._value / x))

    def __gt__(self, other):
        return self._value > other

    def __lt__(self, other):
        return self._value < other

def addPoolLiquidity(amountA,amountB,reserveA,reserveB,totalLiquidity):
    # No Checks included! Only useable for valid block transactions!
    # https://github.com/DeFiCh/ain/blob/a7d914f15f762d121ac2c5d07d38d30cf7e09d4d/src/masternodes/poolpairs.cpp
    if totalLiquidity == 0:
        liq = amountA * amountB
        liquidity = int(math.floor(math.sqrt(liq)))
        if liquidity < MINIMUM_LIQUIDITY:
            return 0
        else:
            return liquidity - MINIMUM_LIQUIDITY
    
    #totalLiquidity = totalLiquidity + MINIMUM_LIQUIDITY
    
    liqA = int(amountA * totalLiquidity / reserveA)
    liqB = int(amountB * totalLiquidity / reserveB)

    return min(liqA, liqB)

def addPoolLiquidityOptimiser():
    # TODO: build opitimiser
    # How to get max liquidity(tokens) from amountA, amountB input
    #########################
    # liquidity -= MINIMUM_LIQUIDITY;
    #     // MINIMUM_LIQUIDITY is a hack for non-zero division
    #     totalLiquidity = MINIMUM_LIQUIDITY;
    ####################################
    # if ((std::max(liqA, liqB) - liquidity) * 100 / liquidity >= 3) {
    #             return Res::Err("Exceeds max ratio slippage protection of 3%%");
    #         }
    pass

def removePoolLiquidity(liquidity,reserveA,reserveB,totalLiquidity):
    # No Checks included! Only useable for valid block transactions!
    # https://github.com/DeFiCh/ain/blob/a7d914f15f762d121ac2c5d07d38d30cf7e09d4d/src/masternodes/poolpairs.cpp
    amountA = int(liquidity * reserveA / totalLiquidity)
    amountB =  int(liquidity * reserveB / totalLiquidity)

    return amountA,amountB

def poolSwap(block,fromAmount,isForward,reserveA,reserveB,commission=0,dexfeeInPct=0,dexfeeOutPct=0):
    # No Checks included! Only useable for valid block transactions!
    # https://github.com/DeFiCh/ain/blob/a7d914f15f762d121ac2c5d07d38d30cf7e09d4d/src/masternodes/poolpairs.cpp -> Swap
    # https://github.com/DeFiCh/ain/blob/d9a26cf042a4a6bb44c824c422990e6c3b4315e8/src/masternodes/mn_checks.cpp -> ExecuteSwap


    # bool const forward = in.nTokenId == idTokenA;
    # auto& reserveF = forward ? reserveA : reserveB;
    # auto& reserveT = forward ? reserveB : reserveA;

    if isForward:
        reserveF = reserveA
        reserveT = reserveB
    else:
        reserveT = reserveA
        reserveF = reserveB

    # claim trading fee
    tradeFee = int(fromAmount * commission)
    fromAmount = fromAmount - tradeFee

    # claim dex fee
    dexfeeInAmount = int(fromAmount * dexfeeInPct)
    fromAmount = fromAmount - dexfeeInAmount

    unswapped = fromAmount
    swapped = 0
    poolFrom = reserveF
    poolTo = reserveT
    if poolFrom/SLOPE_SWAP_RATE < unswapped:
        chunk = poolFrom/SLOPE_SWAP_RATE
    else:
        chunk = unswapped
    if block < BAY_FRONT_GARDENS_HEIGHT:
        while unswapped > 0:
            stepFrom = min(chunk,unswapped)
            stepTo = int(poolTo * stepFrom / poolFrom)
            poolFrom = poolFrom + stepFrom
            poolTo = poolTo - stepTo
            unswapped = unswapped - stepFrom
            swapped = swapped + stepTo
    else:
        # Better results without ArithUint256 (2022/08/11)
        #swapped = (ArithUint256(poolTo) * poolFrom / (poolFrom + unswapped)) * -1 + poolTo
        swapped = poolTo - (poolTo * poolFrom / (poolFrom + unswapped))
        if block >= FORT_CANNING_HILL_HEIGHT and swapped != 0:
            #swapped = math.floor(swapped.value)
            swapped = math.floor(swapped)

        poolFrom = poolFrom + unswapped
        poolTo = poolTo - swapped
    
    dexfeeOutAmount = int(swapped * dexfeeOutPct)
    swapped = swapped - dexfeeOutAmount

    if isForward:
        reserveAChange = poolFrom - reserveF
        reserveBChange = poolTo - reserveT 
    else:
        reserveBChange = poolFrom - reserveF
        reserveAChange = poolTo - reserveT 
    return int(reserveAChange),int(reserveBChange),int(swapped),int(tradeFee),int(dexfeeInAmount),int(dexfeeOutAmount)

#TEST FIRST FOUR DAX-DUSD TX
# if __name__ == '__main__':
# #     amountA = 1000000000
# #     amountB = 1245000000000
# #     reserveA = 0
# #     reserveB = 0
# #     totalLiq = 0
# #     i = addPoolLiquidity(amountA,amountB,reserveA,reserveB,totalLiq)
# #     print(i)

#     reserveA = 444500000000
#     reserveB = 200000000000
#     totalLiq = 917403152
#     amountA = 3500000000
    
#     AChange,BChange,swapped,commission,dexIn,dexOut = poolSwap(480000,amountA,True,reserveA,reserveB,0.002,0,0)
#     print(swapped)
#     reserveA = poolT + 1 #correction
#     reserveB = poolF
#     amountA = 5000000
    
#     poolF,poolT,swapped,dexIn,dexOut = poolSwap(2102378,'DUSD','DAX-DUSD',amountA,reserveA,reserveB,0.002,0.001)
#     print(swapped)
    
#     reserveA = poolT + 2
#     reserveB = poolF
#     amountA = 3300000
#     amountB = 75569367
#     i = addPoolLiquidity(amountA,amountB,reserveA,reserveB,totalLiq)
#     print(i)



import json
import requests
import defichainUtils.utils.bech32 as bech32
import defichainUtils.utils.base58 as base58
# import utils.bech32 as bech32
# import utils.base58 as base58

def int2bytes(i:int, enc, length=None):
    if length == None:
        length = (i.bit_length() + 7) // 8

    return i.to_bytes(length, enc)

def int2hex(i:int, enc, cntBytes=None):
    res = int2bytes(i, enc).hex()
    if cntBytes is not None:
        missingBytes = int(cntBytes - len(res)/2)
        if missingBytes < 0:
            return res
    else:
        missingBytes = 0
    
    if enc == 'little':
        return res + ('00' * missingBytes)
    elif enc == 'big':
        return ('00' * missingBytes) + res
    else:
        return res

def convert_hex(str, enc1, enc2):
    '''
    Convert Hex String from enc1 or enc2: e.g. little (Endian) to big (Endian)
    '''
    return int2bytes(int.from_bytes(bytes.fromhex(str), enc1), enc2).hex()

def hexToString(hex):
    if hex[:2] == '0x':
        hex = hex[2:]
    stringValue = bytes.fromhex(hex).decode('utf-8')
    return stringValue

def stringToHex(string):
    hex = string.encode('utf-8').hex()
    if hex[:2] == '0x':
        hex = hex[2:]
    return hex

def generateSequence(hex:str,locktime=0,replacable=False):
    if replacable:
        reducer = 2
    elif locktime > 0:
        reducer = 1
    else:
        reducer = 0

    return int2hex(int(convert_hex(hex,'little','big'),16) - reducer,'little',4)

def encodeBech32AddressToHex(address):
    '''
    DefiChain related Bech32 Encoded Address to Hex.
    '''
    hrp = address.split('1')[0]
    witnessVersion,witnessProgram = bech32.decode(hrp,address)
    # Address MUST be BECH32: Then witnessVersion and witnessProgramm have 22 Bytes together
    res = f'{22:02x}' + f'{witnessVersion:02x}' + f'{len(witnessProgram):02x}'

    for w in witnessProgram:
        res = res + f'{w:02x}'

    return res

def decodeHexToAddress(hex,addressType,addressVersion):
    if addressType == 'bech32':
        witnessVersion = int(addressVersion,16)
        witnessProgram = []
        addressHex = hex
        for i in range(0,int(len(addressHex)/2)):
            witnessProgram.append(int(addressHex[i*2:i*2+2],16))
        #witnessProgram = hex[4:]

        return bech32.encode('df',witnessVersion,witnessProgram)
    elif addressType == 'base58':
        # Add Version 90 (https://bitcoin.stackexchange.com/questions/36623/what-version-byte-will-produce-a-base58-encoding-starting-with-x) Adds "d" to the beginning
        return base58.b58encode_chk(bytes.fromhex(int2hex(addressVersion,'little') + hex))
    else:
        return None

def getTokenId(token:str, rpcConnector):
    ''' 
    From dToken name (e.g. KO) to dToken ID (e.g. 74)
    '''
    res = rpc(rpcConnector,'gettoken',[token])   
    if res is None:
        return None
    return int(list(res.keys())[0])
        
# RPC
def rpc(rpcConnector, method, params=None, silentErrors=False):
    if params[0] == 'DUSD':
        id = 11
    else:
        id = 12
    return {
        id: {
            "bla": 1
        }
    }
    '''
    calling RPC Methos. rpcConnector must be RPCConnector Class
    '''
    if params is None:
        params = []
    data = json.dumps({
        "jsonrpc": "2.0",
        "id": "meBe",
        "method": method,
        "params": params
    })
    result = requests.post(rpcConnector.NODE_URL, auth=(rpcConnector.NODE_USER, rpcConnector.NODE_PASSWORD), data=data)
    if result.status_code >= 300 and not silentErrors:
        return None
    return result.json()['result']


# import utils.utils as utils
# from utils.CustomTx import CustomTxType
import defichainUtils.utils.utils as utils
from defichainUtils.utils.CustomTx import CustomTxType
from dataclasses import asdict


def decodeCustomTx(hex,DfMarker):
    '''
    Hex String without OP_RETURN and DfTx Marker
    ''' 
    DfMarker = utils.hexToString(DfMarker)
    if DfMarker == 'DfTx':
        functionMarker = utils.hexToString(hex[:2])
        if functionMarker == 'j':
            res = decodeSetGovVariable(hex[2:],True)
            if res != None:
                res['name'] = 'setgovheight'
        elif functionMarker == 'G':
            res = decodeSetGovVariable(hex[2:],False)
            if res != None:
                res['name'] = 'setgovvariable'
        elif functionMarker == 's':
            res = decodePoolSwap(hex[2:])
            if res != None:
                res['name'] = 'poolswap'
        elif functionMarker == 'i':
            res = decodePoolSwapV2(hex[2:])
            if res != None:
                res['name'] = 'compositeswap'
        elif functionMarker == 'Q':
            res = decodeDFIP2203(hex[2:])
            if res != None:
                res['name'] = 'dfip2203'
        elif functionMarker == 'l':
            res = decodeAddPoolLiquidity(hex[2:])
            if res != None:
                res['name'] = 'addpoolliquidity'
        elif functionMarker == 'r':
            res = decodeRemovePoolLiquidity(hex[2:])
            if res != None:
                res['name'] = 'removepoolliquidity'
        elif functionMarker == 'p':
            res = decodeCreatePoolPair(hex[2:])
            if res != None:
                res['name'] = 'createpoolpair'
        elif functionMarker == 'B':
            res = decodeAccountToAccount(hex[2:])
            if res != None:
                res['name'] = 'accounttoaccount'
        elif functionMarker == 'X':
            res = decodeTakeLoan(hex[2:])
            if res != None:
                res['name'] = 'takeloan'
        elif functionMarker == 'H':
            res = decodePaybackLoan(hex[2:])
            if res != None:
                res['name'] = 'paybackloan'
        elif functionMarker == 'k':
            res = decodePaybackLoanV2(hex[2:])
            if res != None:
                res['name'] = 'paybackloanv2'
        elif functionMarker == 'a':
            res = decodeAnyAccountsToAccounts(hex[2:])
            if res != None:
                res['name'] = 'anyaccountstoaccounts'
        elif functionMarker == 'e':
            res = decodeCloseVault(hex[2:])
            if res != None:
                res['name'] = 'closevault'
        elif functionMarker == 'V':
            res = decodeVault(hex[2:])
            if res != None:
                res['name'] = 'vault'
        elif functionMarker == 'v':
            res = decodeUpdateVault(hex[2:])
            if res != None:
                res['name'] = 'updatevault'
        elif functionMarker == 'I':
            res = decodeAuctionBid(hex[2:])
            if res != None:
                res['name'] = 'auctionbid'
        elif functionMarker == 'J':
            res = decodeWithdrawFromVault(hex[2:])
            if res != None:
                res['name'] = 'withdrawfromvault'
        elif functionMarker == 'S':
            res = decodeDepositToVault(hex[2:])
            if res != None:
                res['name'] = 'deposittovault'          
        else:
            c = CustomTxType()
            for key, value in asdict(c).items():
                if value == functionMarker:
                    return {
                        'name': str(key).lower(),
                        'hex': hex[2:]
                    }
            return None
        return res
    elif DfMarker == 'DfTS':
        res = decodeDfTxTokenSplit(hex)
        if res != None:
            res['name'] = 'DfTS'
        return res
    else:
        return {
            'txType': DfMarker,
            'hex': hex
        }


def decodeSetGovVariable(hex,withHeight):
    # Only data part of hex string!
    govVariable = hex[2:2+2*int(hex[:2],16)]
    hex = hex[2+2*int(hex[:2],16):]
    if utils.stringToHex('ATTRIBUTES') == govVariable:
        cnt = int(hex[:2],16)
        hex = hex[2:]
        result = {
            'govvariable': 'ATTRIBUTES',
            'attributes': [],
        }
        for i in range(0,cnt):
            typeMarker = utils.hexToString(hex[8:10])
            if typeMarker in ['p']:
                variables = [
                    {'bytesCount': 1, 'key': 'type', 'type': str,'endian':None,'abort':['o']},
                    {'bytesCount': 1, 'key': 'typeId', 'type': int,'endian':None},
                    {'bytesCount': 4, 'key': 'key', 'type': str,'endian':'big'},
                    {'bytesCount': 4, 'key': 'keyId', 'type': int,'endian':'little'},
                    {'bytesCount': 'rest', 'key': 'value', 'type': int,'endian':'little'}
                ]
                result['attributes'].append(searchHex(variables,hex[8:44]))
                hex = hex[44:]
            elif typeMarker in ['L']:
                variables = [
                    {'bytesCount': 1, 'key': 'type', 'type': str,'endian':None,'abort':['o']},
                    {'bytesCount': 1, 'key': 'key', 'type': str,'endian':None},
                    {'bytesCount': 4, 'key': 'keyId', 'type': int,'endian':None},
                    {'bytesCount': 4, 'key': 'unknown1', 'type': int,'endian':'little'},
                    {'bytesCount': 1, 'key': 'unknown2', 'type': int,'endian':'little'}
                ]
                result['attributes'].append(searchHex(variables,hex[8:30]))
                hex = hex[30:]
            elif typeMarker in ['o']: 
                variables = [
                    {'bytesCount': 1, 'key': 'type', 'type': str,'endian':None},
                    {'bytesCount': 1, 'key': 'typeId', 'type': str,'endian':None},
                    {'bytesCount': 7, 'key': 'unkown', 'type': 'hex','endian':None},
                    {'bytesCount': 4, 'key': 'splitType', 'type': int,'endian':None},
                    {'bytesCount': 4, 'key': 'tokenId', 'type': int,'endian':'little'},
                    {'bytesCount': 4, 'key': 'multiplier', 'type': int,'endian':'little'}
                ]
                result['attributes'].append(searchHex(variables,hex[8:50]))
                hex = hex[50:]

        if withHeight: 
            result['height'] = int(utils.convert_hex(hex[-8:],'little','big'),16)
            #{'bytesCount': 4, 'key': 'height', 'type': int,'endian':'little'}
        
        # variables = [
        #     {'key':'govvariable', 'bytesCount':0, 'type': 'constant', 'value': 'ATTRIBUTES','endian':None},
        #     {'key':'attributes', 'type': [
        #         {'bytesCount': 4, 'key': 'unknown', 'type': int,'endian':None},
        #         {'bytesCount': 1, 'key': 'type', 'type': str,'endian':None,'abort':['o']},
        #         {'bytesCount': 1, 'key': 'typeId', 'type': int,'endian':None},
        #         {'bytesCount': 4, 'key': 'key', 'type': str,'endian':'big'},
        #         {'bytesCount': 4, 'key': 'keyId', 'type': int,'endian':'little'},
        #         {'bytesCount': 8, 'key': 'value', 'type': int,'endian':'little'}]
        #     },
        #     {'bytesCount': 4, 'key': 'height', 'type': int,'endian':'little'}
        # ]
        return result 
    elif utils.stringToHex('ICX_TAKERFEE_PER_BTC') == govVariable:
        return {
            'govvariable': 'ICX_TAKERFEE_PER_BTC',
            'hex': hex
        }
    elif utils.stringToHex('LP_LOAN_TOKEN_SPLITS') == govVariable:
        return {
            'govvariable': 'LP_LOAN_TOKEN_SPLITS',
            'hex': hex
        }
    elif utils.stringToHex('LP_SPLITS') == govVariable:
        return {
            'govvariable': 'LP_SPLITS',
            'hex': hex
        }
    elif utils.stringToHex('ORACLE_DEVIATION') == govVariable:
        return {
            'govvariable': 'ORACLE_DEVIATION',
            'hex': hex
        }
    else:
        return None

def decodePoolSwap(hex):
    variables = [
        {'key': 'from', 'type': 'address','endian':None},
        {'bytesCount':1, 'key': 'fromTokenId', 'type': int,'endian':None},
        {'bytesCount': 8, 'key': 'fromAmount', 'type': int,'endian':'little'},
        {'key': 'to', 'type': 'address','endian':None},
        {'bytesCount':1, 'key': 'toTokenId', 'type': int,'endian':None},
        {'bytesCount': 8, 'key': 'maxPriceCounter', 'type': int,'endian':'little'},
        {'bytesCount': 8, 'key': 'maxPriceDenominator', 'type': int,'endian':'little'}
    ]
    return searchHex(variables,hex)
def decodePoolSwapV2(hex):
    variables = [
        {'key': 'from', 'type': 'address','endian':None},
        {'bytesCount':1, 'key': 'fromTokenId', 'type': int,'endian':None},
        {'bytesCount': 8, 'key': 'fromAmount', 'type': int,'endian':'little'},
        {'key': 'to', 'type': 'address','endian':None},
        {'bytesCount':1, 'key': 'toTokenId', 'type': int,'endian':None},
        {'bytesCount': 8, 'key': 'maxPriceCounter', 'type': int,'endian':'little'},
        {'bytesCount': 8, 'key': 'maxPriceDenominator', 'type': int,'endian':'little'},
        {'key':'pools', 'type':
            [{'bytesCount': 1, 'type': int,'endian':None}]
        }
    ]
    return searchHex(variables,hex)
def decodeDFIP2203(hex):
    variables = [
        {'key': 'from', 'type': 'address','endian':None},
        {'bytesCount':1, 'key': 'fromTokenId', 'type': int,'endian':None},
        {'bytesCount': 8, 'key': 'fromAmount', 'type': int,'endian':'little'},
        {'bytesCount':1, 'key': 'toTokenId', 'type': int,'endian':None},
        {'bytesCount':4, 'key': 'direction', 'type': int,'endian':None}
    ]
    return searchHex(variables,hex)
    
def decodeAddPoolLiquidity(hex):
    variables = [
        {'key':'addresses', 'type': [
            {'key': 'from', 'type': 'address','endian':None},
            {'key': 'amounts', 'type': [
                {'bytesCount': 4,'key': 'tokenId', 'type': int,'endian':'little'},
                {'bytesCount': 8, 'key': 'amount', 'type': int,'endian':'little'}
                ]}
        ]},
        {'key': 'shareAddress', 'type': 'address','endian':None}
    ]
    return searchHex(variables,hex)
def decodeRemovePoolLiquidity(hex):
    variables = [
        {'key': 'from', 'type': 'address','endian':None},
        {'bytesCount':1, 'key': 'tokenId', 'type': int,'endian':None},
        {'bytesCount': 8, 'key': 'amount', 'type': int,'endian':'little'}
    ]
    return searchHex(variables,hex)
def decodeCreatePoolPair(hex):
    variables = [
        {'bytesCount':1, 'key': 'tokenIdA', 'type': int,'endian':None},
        {'bytesCount':1, 'key': 'tokenIdB', 'type': int,'endian':None},
        {'bytesCount': 8, 'key': 'commission', 'type': int,'endian':'little'},
        {'key': 'from', 'type': 'address','endian':None},
        {'bytesCount':1, 'key': 'status', 'type': int,'endian':None},
        {'key': 'customRewards', 'type': 'rest' ,'endian':None}
    ]
    return searchHex(variables,hex)
def decodeAccountToAccount(hex):
    variables = [
        {'key': 'from', 'type': 'address','endian':None},
        {'key':'addresses', 'type': [
            {'key': 'to', 'type': 'address','endian':None},
            {'key': 'amounts', 'type': [
                {'bytesCount': 4,'key': 'tokenId', 'type': int,'endian':'little'},
                {'bytesCount': 8, 'key': 'amount', 'type': int,'endian':'little'}
                ]}
        ]}
    ]
    return searchHex(variables,hex)
def decodeTakeLoan(hex):
    variables = [
        {'bytesCount': 32,'key': 'vaultId', 'type': 'hex','endian':'little', 'keepLeadingZeros': True},
        {'key': 'to', 'type': 'address','endian':None},
        {'key': 'amounts', 'type': [
            {'bytesCount': 4,'key': 'tokenId', 'type': int,'endian':'little'},
            {'bytesCount': 8, 'key': 'amount', 'type': int,'endian':'little'}
        ]}
    ]
    return searchHex(variables,hex)
def decodePaybackLoan(hex):
    variables = [
        {'bytesCount': 32,'key': 'vaultId', 'type': 'hex','endian':'little', 'keepLeadingZeros': True},
        {'key': 'from', 'type': 'address','endian':None},
        {'key': 'amounts', 'type': [
            {'bytesCount': 4,'key': 'tokenId', 'type': int,'endian':'little'},
            {'bytesCount': 8, 'key': 'amount', 'type': int,'endian':'little'}
        ]}
    ]
    return searchHex(variables,hex)
def decodePaybackLoanV2(hex):
    variables = [
        {'bytesCount': 32,'key': 'vaultId', 'type': 'hex','endian':'little', 'keepLeadingZeros': True},
        {'key': 'from', 'type': 'address','endian':None},
        {'key': 'tokens', 'type': [
            {'bytesCount': 1,'key': 'tokenId', 'type': int,'endian':None},
            {'key': 'amounts', 'type': [
                {'bytesCount': 4,'key': 'tokenId', 'type': int,'endian':'little'},
                {'bytesCount': 8, 'key': 'amount', 'type': int,'endian':'little'}
            ]}
        ]}
    ]
    return searchHex(variables,hex)
def decodeAnyAccountsToAccounts(hex):
    variables = [
        {'key':'fromAddresses', 'type': [
            {'key': 'from', 'type': 'address','endian':None},
            {'key': 'amounts', 'type': [
                {'bytesCount': 4,'key': 'tokenId', 'type': int,'endian':'little'},
                {'bytesCount': 8, 'key': 'amount', 'type': int,'endian':'little'}
                ]}
        ]},
        {'key':'toAddresses', 'type': [
            {'key': 'to', 'type': 'address','endian':None},
            {'key': 'amounts', 'type': [
                {'bytesCount': 4,'key': 'tokenId', 'type': int,'endian':'little'},
                {'bytesCount': 8, 'key': 'amount', 'type': int,'endian':'little'}
                ]}
        ]}
    ]
    return searchHex(variables,hex)
def decodeCloseVault(hex):
    variables = [
        {'bytesCount': 32, 'key': 'vaultId', 'type': 'hex', 'endian':'little', 'keepLeadingZeros': True},
        {'key': 'shareAddress', 'type': 'address','endian':None}
    ]
    return searchHex(variables,hex)
def decodeUpdateVault(hex):
    variables = [
        {'bytesCount': 32, 'key': 'vaultId', 'type': 'hex', 'endian':'little', 'keepLeadingZeros': True},
        {'key': 'ownerAddress', 'type': 'address','endian':None},
        {'bytesCount': 1,'key': 'lengthLoanSchemeName', 'type': int,'endian':'little'},
        {'bytesCount': 'rest', 'key': 'loanScheme', 'type': str,'endian':None}
    ]
    return searchHex(variables,hex)
def decodeVault(hex):
    variables = [
        {'key': 'address', 'type': 'address','endian':None},
        {'bytesCount': 1,'key': 'lengthLoanSchemeName', 'type': int,'endian':'little'},
        {'bytesCount': 'rest', 'key': 'loanScheme', 'type': str,'endian':None}
    ]
    return searchHex(variables,hex)
def decodeAuctionBid(hex):
    variables = [
        {'bytesCount': 32, 'key': 'vaultId', 'type': 'hex', 'endian':'little', 'keepLeadingZeros': True},
        {'bytesCount': 4,'key': 'index', 'type': int,'endian':'little'},
        {'key': 'address', 'type': 'address','endian':None},
        {'bytesCount': 1, 'key': 'tokenId', 'type': int,'endian':'little'},
        {'bytesCount': 8,'key': 'amount', 'type': int,'endian':'little'}
    ]
    return searchHex(variables,hex)
def decodeWithdrawFromVault(hex):
    variables = [
        {'bytesCount': 32, 'key': 'vaultId', 'type': 'hex', 'endian':'little', 'keepLeadingZeros': True},
        {'key': 'to', 'type': 'address','endian':None},
        {'bytesCount': 1, 'key': 'tokenId', 'type': int,'endian':'little'},
        {'bytesCount': 8,'key': 'amount', 'type': int,'endian':'little'}
    ]
    return searchHex(variables,hex)
def decodeDepositToVault(hex):
    variables = [
        {'bytesCount': 32, 'key': 'vaultId', 'type': 'hex', 'endian':'little', 'keepLeadingZeros': True},
        {'key': 'from', 'type': 'address','endian':None},
        {'bytesCount': 1, 'key': 'tokenId', 'type': int,'endian':'little'},
        {'bytesCount': 8,'key': 'amount', 'type': int,'endian':'little'}
    ]
    return searchHex(variables,hex)

# NOT a DfTx Function!
def decodeDfTxTokenSplit(hex):
    variables = [
        {'bytesCount': 4, 'key': 'typeId', 'type': int,'endian':'little'},
        {'bytesCount': 4,'key': 'tokenId', 'type': int,'endian':'little'},
        {'bytesCount': 4,'key': 'multiplier', 'type': int,'endian':'little'}
    ]
    return searchHex(variables,hex)
    
def recursiveSearchHex(variable,hex):
    abort = False
    if type(variable['type']) == list:
        cnt = int(hex[:2],16)
        hex = hex[2:]
        arr = []
        for i in range(0,cnt):
            res = {}
            for idx,listVariable in enumerate(variable['type']):
                (key,value,hex,abort) = recursiveSearchHex(listVariable,hex)
                if 'abort' in variable['type'][idx] and value in variable['type'][idx]['abort']:
                    abort = True
                    break
                if key is None and idx == len(variable['type']) - 1:
                    res = value
                else:
                    res[key] = value
            arr.append(res)
        return variable['key'],arr,hex,abort
    else:
        (key,value,hex) = getValue(variable,hex)
        return key,value,hex,abort

def searchHex(variables:list,hex:str):
    res = {}
    for variable in variables:
        (key,value,hex,abort) = recursiveSearchHex(variable,hex)
        res[key] = value
        if abort:
            break
    
    if len(hex) == 0:
        res['hex'] = None
    else:
        res['hex'] = hex
    return res

def getValue(variable,hex):
    # Select subString from hex depending on length info in hex String or length info from variable input
    if variable['type'] == 'address':
        if 'key' in variable and variable['key'] == 'to' and hex[:2] == '00':
            return variable['key'],None,hex[2:]
        # START - Remove leading zeros
        while True:
            if len(hex) >= 2 and hex[:2] == '00':
                hex = hex[2:]
            else:
                break
        if len(hex) < 2:
            return None,None,None
        # END - Remove leading zeros
        hexLength = int(hex[:2],16)
        if hex[2:6] == '76a9' and hexLength == 25:
            addressType = 'base58'
            addressVersion = 18
            addressLength = int(hex[6:8],16)
            hexOfInterest = hex[8:8+2*addressLength]
            hex = hex[2+2*hexLength:]
        elif hex[2:4] == 'a9' and hexLength == 23: 
            addressType = 'base58'
            addressVersion = 90
            addressLength = int(hex[4:6],16)
            hexOfInterest = hex[6:6+2*addressLength]
            hex = hex[2+2*hexLength:]
        else:
            addressType = 'bech32'
            addressVersion = hex[2:4]
            addressLength = int(hex[4:6],16)
            hexOfInterest = hex[6:6+2*addressLength]
            hex = hex[2+2*hexLength:]
    elif variable['type'] == 'rest':
        hexOfInterest = hex
        hex = ''
    else:
        if variable['bytesCount'] == 'rest':
            hexOfInterest = hex
            hex = ''
        elif variable['bytesCount'] == 0:
            hexOfInterest = ''
        else:
            hexOfInterest = hex[:2*variable['bytesCount']]
            hex = hex[2*variable['bytesCount']:]

    # Handle Endian Transformation
    hexZeros = ''
    if 'keepLeadingZeros' in variable and variable['keepLeadingZeros'] and hexOfInterest[-2:] == '00':
        counter = 0
        while True:
            search = hexOfInterest[- 2 * (counter + 1):] if counter == 0 else hexOfInterest[- 2 * (counter + 1):-2 * counter]
            if search == '00':
                hexZeros += '00'
                counter += 1
            else:
                break
    if variable['endian'] == 'big':
        hexOfInterest = hexZeros + utils.convert_hex(hexOfInterest,'big','little')
    elif variable['endian'] == 'little':
        hexOfInterest = hexZeros + utils.convert_hex(hexOfInterest,'little','big')

    # Extract Value
    if variable['type'] == int:
        if hexOfInterest == '':
            valueOfInterest = 0
        else:
            valueOfInterest = int(hexOfInterest,16)
    elif variable['type'] == str:
        if hexOfInterest == '':
            valueOfInterest = hexOfInterest
        else:
            valueOfInterest = utils.hexToString(hexOfInterest)
    elif variable['type'] == 'address':
        valueOfInterest = utils.decodeHexToAddress(hexOfInterest,addressType,addressVersion)
    elif variable['type'] == 'hex':
        valueOfInterest = hexOfInterest
    elif variable['type'] == 'constant':
        valueOfInterest = variable['value']
    else:
        valueOfInterest = None
    
    if 'key' not in variable:
        return None,valueOfInterest,hex
    if variable['key'] != None:
        return variable['key'],valueOfInterest,hex
    return None,None,hex


import defichainUtils.utils.utils as utils
# import utils.utils as utils
from dataclasses import dataclass, is_dataclass



DECIMALS = 100000000
class DfTx():
    def __init__(self):
        self.params = []
        self.DfTxMarker = 'DfTx'
    
    def hex(self):
        h = self.DfTxMarker.encode('utf-8').hex()
        lengthInfo = ''
        for i,p in enumerate(self.params):
            h = h + self.__handleParam(p,i==0)
            # TODO: This part is not understood yet. It is a workaround. Some Functions have 4c before the Byte handling the length. Some not. Semms to depend on length. 
            if i == 0 and p not in [CustomTxType.DepositToVault]:
                lengthInfo = '4c'

        return OPCodes.OP_RETURN + lengthInfo + utils.int2hex(int(len(h)/2),'little') + h

    def __handleParam(self,p,isCustomTxType=False):
        if isCustomTxType:
            return p.encode('utf-8').hex()
        elif not is_dataclass(p):
            return p
        
        
        if p.__class__.__name__ == 'Vault':
            return utils.convert_hex(p.value,'big','little')
        elif p.__class__.__name__ == 'Address':
            return utils.encodeBech32AddressToHex(p.value)
        # elif p.__class__.__name__ == 'Addresses':
        #     h = ''
        #     for address in p.value:
        #         h = h + encodeBech32AddressToHex(address)
        #     return int2hex(len(p.value),'little') + h
        elif p.__class__.__name__ == 'Amounts':
            h = ''
            for amount in p.value:
                h = h + utils.int2hex(amount[1],'little',4) + utils.int2hex(int(amount[0]*DECIMALS),'little',8)
            return utils.int2hex(len(p.value),'little') + h
        elif p.__class__.__name__ == 'AddressesWithAmount':
            d = p.toDict()
            h = ''
            for address in d:
                # 01 means, count of addresses (=1)
                h = h + '01' + utils.encodeBech32AddressToHex(address) + utils.int2hex(len(d[address].keys()),'little')
                for tokenId in d[address]:
                    h = h + utils.int2hex(tokenId,'little',4) + utils.int2hex(int(d[address][tokenId]*DECIMALS),'little',8)
            return h
        else:
            return ''

@dataclass
class CustomTxType:
    Reject:str = 1
    # masternodes:
    CreateMasternode:str = 'C'
    ResignMasternode:str = 'R'
    UpdateMasternode:str = 'm'
    SetForcedRewardAddress:str = 'F'
    RemForcedRewardAddress:str = 'f'
    
    # custom tokens:
    CreateToken:str = 'T'
    MintToken:str   = 'M'
    UpdateToken:str = 'N' # previous type, only DAT flag triggers
    UpdateTokenAny:str = 'n' # new type of token's update with any flags/fields possible
    
    # dex orders - just not to overlap in future
    # CreateOrder:str = 'O'
    # DestroyOrder:str = 'E'
    # MatchOrders:str = 'A'
    
    # poolpair
    CreatePoolPair:str = 'p'
    UpdatePoolPair:str = 'u'
    PoolSwap:str = 's'
    PoolSwapV2:str = 'i'
    AddPoolLiquidity:str = 'l'
    RemovePoolLiquidity:str = 'r'
    
    # accounts
    UtxosToAccount:str = 'U'
    AccountToUtxos:str = 'b'
    AccountToAccount:str = 'B'
    AnyAccountsToAccounts:str = 'a'
    SmartContract:str = 'K'
    DFIP2203:str = 'Q'
    # set governance variable
    SetGovVariable:str = 'G'
    SetGovVariableHeight:str = 'j'
    # Auto auth TX
    AutoAuthPrep:str  = 'A'
    # oracles
    AppointOracle:str = 'o'
    RemoveOracleAppoint:str = 'h'
    UpdateOracleAppoint:str = 't'
    SetOracleData:str = 'y'
    # ICX
    ICXCreateOrder:str = '1'
    ICXMakeOffer:str = '2'
    ICXSubmitDFCHTLC:str = '3'
    ICXSubmitEXTHTLC:str = '4'
    ICXClaimDFCHTLC:str = '5'
    ICXCloseOrder:str = '6'
    ICXCloseOffer:str = '7'
    
    # Loans
    SetLoanCollateralToken:str = 'c'
    SetLoanToken:str = 'g'
    UpdateLoanToken:str = 'x'
    LoanScheme:str = 'L'
    DefaultLoanScheme:str = 'd'
    DestroyLoanScheme:str = 'D'
    Vault:str = 'V'
    CloseVault:str = 'e'
    UpdateVault:str = 'v'
    DepositToVault:str = 'S'
    WithdrawFromVault:str = 'J'
    TakeLoan:str = 'X'
    PaybackLoan:str = 'H'
    PaybackLoanV2:str = 'k'
    AuctionBid:str = 'I'
    
    # Marker TXs
    FutureSwapExecution:str = 'q'
    FutureSwapRefund:str = 'w'

@dataclass
class OPCodes:
    OP_RETURN:str = '6a'